@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2008 - 2021, Ilan Schnell
+   Copyright (c) 2008 - 2021, Ilan Schnell; All Rights Reserved
    bitarray is published under the PSF license.
 
    This file is the C part of the bitarray package.
@@ -14,7 +14,7 @@
 #include "bitarray.h"
 #include "bitarray_vct.h"
 
-/* block size used when reading / writing blocks of bytes from files */
+/* size used when reading / writing blocks from files (in bytes) */
 #define BLOCKSIZE  65536
 
 #ifdef IS_PY3K
@@ -140,8 +140,7 @@ newbitarrayobject(PyTypeObject *type, Py_ssize_t nbits, int endian)
         obj->ob_item = (char *) PyMem_Malloc((size_t) nbytes);
         if (obj->ob_item == NULL) {
             PyObject_Del(obj);
-            PyErr_NoMemory();
-            return NULL;
+            return PyErr_NoMemory();
         }
     }
     obj->allocated = nbytes;
@@ -164,6 +163,37 @@ bitarray_dealloc(bitarrayobject *self)
     Py_TYPE(self)->tp_free((PyObject *) self);
 }
 
+/* reverse n bytes (starting at start) in buffer */
+static void
+bytereverse(bitarrayobject *self, Py_ssize_t start, Py_ssize_t n)
+{
+    static char trans[256];
+    static int setup = 0;
+    Py_ssize_t i, stop = start + n;
+
+    assert(n >= 0);
+    assert(0 <= start && start <= Py_SIZE(self));
+    assert(0 <= stop && stop <= Py_SIZE(self));
+
+    if (!setup) {
+        /* setup translation table, which maps each byte to it's reversed:
+           trans = {0, 128, 64, 192, 32, 160, ..., 255} */
+        int j, k;
+
+        for (k = 0; k < 256; k++) {
+            trans[k] = 0x00;
+            for (j = 0; j < 8; j++)
+                if (1 << (7 - j) & k)
+                    trans[k] |= 1 << j;
+        }
+        setup = 1;
+    }
+
+    setunused(self);
+    for (i = start; i < stop; i++)
+        self->ob_item[i] = trans[(unsigned char) self->ob_item[i]];
+}
+
 /* copy n bits from other (starting at b) onto self (starting at a) */
 static void
 copy_n(bitarrayobject *self, Py_ssize_t a,
@@ -181,20 +211,21 @@ copy_n(bitarrayobject *self, Py_ssize_t a,
        bytes using memmove, and copy the remaining few bits individually.
        Note that the order of these two operations matters when copying
        self to self. */
-    if (self->endian == other->endian && a % 8 == 0 && b % 8 == 0 && n >= 8)
+    if (a % 8 == 0 && b % 8 == 0 && n >= 8)
     {
         const size_t bytes = n / 8;
         const Py_ssize_t bits = BITS(bytes);
 
-        assert(bits <= n && n < bits + 8);
-        if (a <= b)
-            memmove(self->ob_item + a / 8, other->ob_item + b / 8, bytes);
-
-        if (n != bits)
+        assert(bytes > 0 && bits <= n && n < bits + 8);
+        if (a > b)
             copy_n(self, bits + a, other, bits + b, n - bits);
 
-        if (a > b)
-            memmove(self->ob_item + a / 8, other->ob_item + b / 8, bytes);
+        memmove(self->ob_item + a / 8, other->ob_item + b / 8, bytes);
+        if (self->endian != other->endian)
+            bytereverse(self, a / 8, bytes);
+
+        if (a <= b)
+            copy_n(self, bits + a, other, bits + b, n - bits);
 
         return;
     }
@@ -454,7 +485,7 @@ extend_iter(bitarrayobject *self, PyObject *iter)
     PyObject *item;
 
     assert(PyIter_Check(iter));
-    while ((item = PyIter_Next(iter)) != NULL) {
+    while ((item = PyIter_Next(iter))) {
         if (append_item(self, item) < 0) {
             Py_DECREF(item);
             return -1;
@@ -492,65 +523,62 @@ extend_list(bitarrayobject *self, PyObject *list)
 }
 
 static int
-extend_tuple(bitarrayobject *self, PyObject *tuple)
+extend_bytes01(bitarrayobject *self, PyObject *bytes)
 {
-    PyObject *item;
-    Py_ssize_t n, i;
+    const Py_ssize_t original_nbits = self->nbits;
+    unsigned char c;
+    char *data;
+    int vi = 0;  /* to avoid uninitialized warning for some compilers */
 
-    assert(PyTuple_Check(tuple));
-    n = PyTuple_GET_SIZE(tuple);
-    if (n == 0)
-        return 0;
+    assert(PyBytes_Check(bytes));
+    data = PyBytes_AS_STRING(bytes);
 
-    if (resize(self, self->nbits + n) < 0)
-        return -1;
-
-    for (i = 0; i < n; i++) {
-        item = PyTuple_GET_ITEM(tuple, i);
-        if (item == NULL)
+    while ((c = *data++)) {
+        switch (c) {
+        case '0': vi = 0; break;
+        case '1': vi = 1; break;
+        case ' ':
+        case '\n':
+        case '\r':
+        case '\t':
+        case '\v':
+            continue;
+        default:
+            /* ensure no bits are added on error */
+            if (resize(self, original_nbits) < 0)
+                return -1;
+            PyErr_Format(PyExc_ValueError, "expected '0' or '1' "
+                         "(or whitespace), got '%c' (0x%02x)", c, c);
             return -1;
-        if (set_item(self, self->nbits - n + i, item) < 0)
+        }
+        if (resize(self, self->nbits + 1) < 0)
             return -1;
+        setbit(self, self->nbits - 1, vi);
     }
     return 0;
 }
 
 static int
-extend_01(bitarrayobject *self, PyObject *bytes)
+extend_unicode01(bitarrayobject *self, PyObject *unicode)
 {
-    Py_ssize_t nbytes, i;
-    char c, *data;
-    int vi = 0;
+    PyObject *bytes;
+    int res;
 
-    assert(PyBytes_Check(bytes));
-    nbytes = PyBytes_GET_SIZE(bytes);
-    if (nbytes == 0)
-        return 0;
-
-    if (resize(self, self->nbits + nbytes) < 0)
+    assert(PyUnicode_Check(unicode));
+    bytes = PyUnicode_AsASCIIString(unicode);
+    if (bytes == NULL)
         return -1;
 
-    data = PyBytes_AsString(bytes);
-    for (i = 0; i < nbytes; i++) {
-        c = data[i];
-        switch (c) {
-        case '0': vi = 0; break;
-        case '1': vi = 1; break;
-        default:
-            PyErr_Format(PyExc_ValueError,
-                         "character must be '0' or '1', found '%c'", c);
-            return -1;
-        }
-        setbit(self, self->nbits - nbytes + i, vi);
-    }
-    return 0;
+    assert(PyBytes_Check(bytes));
+    res = extend_bytes01(self, bytes);
+    Py_DECREF(bytes);  /* drop bytes */
+    return res;
 }
 
 static int
 extend_dispatch(bitarrayobject *self, PyObject *obj)
 {
     PyObject *iter;
-    int res;
 
     /* dispatch on type */
     if (bitarray_Check(obj))                              /* bitarray */
@@ -559,32 +587,19 @@ extend_dispatch(bitarrayobject *self, PyObject *obj)
     if (PyList_Check(obj))                                    /* list */
         return extend_list(self, obj);
 
-    if (PyTuple_Check(obj))                                  /* tuple */
-        return extend_tuple(self, obj);
-
     if (PyBytes_Check(obj)) {                             /* bytes 01 */
-        /* This case is used on Python 2.  However, it should have never
-           been here for Python 3, as it allows bitarray(b'01101011') */
 #ifdef IS_PY3K
-        if (PyErr_WarnEx(PyExc_DeprecationWarning,
-                         "cannot extend from 'bytes', "
-                         "use .pack() or .frombytes() instead", 1) < 0)
-            return -1;
+        PyErr_SetString(PyExc_TypeError,
+                        "cannot extend bitarray with 'bytes', "
+                        "use .pack() or .frombytes() instead");
+        return -1;
+#else
+        return extend_bytes01(self, obj);
 #endif
-        return extend_01(self, obj);
     }
 
-    if (PyUnicode_Check(obj)) {                /* (unicode) string 01 */
-        PyObject *bytes;
-
-        bytes = PyUnicode_AsEncodedString(obj, NULL, NULL);
-        if (bytes == NULL)
-            return -1;
-        assert(PyBytes_Check(bytes));
-        res = extend_01(self, bytes);
-        Py_DECREF(bytes);  /* drop bytes */
-        return res;
-    }
+    if (PyUnicode_Check(obj))                           /* unicode 01 */
+        return extend_unicode01(self, obj);
 
     if (PyIter_Check(obj))                                    /* iter */
         return extend_iter(self, obj);
@@ -592,10 +607,13 @@ extend_dispatch(bitarrayobject *self, PyObject *obj)
     /* finally, try to get the iterator of the object */
     iter = PyObject_GetIter(obj);
     if (iter) {
+        int res;
+
         res = extend_iter(self, iter);
         Py_DECREF(iter);
         return res;
     }
+
     PyErr_Format(PyExc_TypeError,
                  "'%s' object is not iterable", Py_TYPE(obj)->tp_name);
     return -1;
@@ -608,16 +626,10 @@ unpack(bitarrayobject *self, char zero, char one, const char *fmt)
     Py_ssize_t i;
     char *str;
 
-    if (self->nbits > PY_SSIZE_T_MAX / 8) {
-        PyErr_SetString(PyExc_OverflowError, "bitarray too large to unpack");
-        return NULL;
-    }
-
     str = (char *) PyMem_Malloc((size_t) self->nbits);
-    if (str == NULL) {
-        PyErr_NoMemory();
-        return NULL;
-    }
+    if (str == NULL)
+        return PyErr_NoMemory();
+
     for (i = 0; i < self->nbits; i++)
         str[i] = GETBIT(self, i) ? one : zero;
 
@@ -655,12 +667,12 @@ IntBool_AsInt(PyObject *v)
         x = PyLong_AsLong(v);
     }
     else {
-        PyErr_SetString(PyExc_TypeError, "integer or bool expected");
+        PyErr_SetString(PyExc_TypeError, "int or bool expected");
         return -1;
     }
 
     if (x < 0 || x > 1) {
-        PyErr_SetString(PyExc_ValueError, "integer 0 or 1 expected");
+        PyErr_SetString(PyExc_ValueError, "int 0 or 1 expected");
         return -1;
     }
     return (int) x;
@@ -765,10 +777,9 @@ bitarray_index(bitarrayobject *self, PyObject *args)
     normalize_index(self->nbits, &stop);
 
     i = findfirst(self, vi, start, stop);
-    if (i < 0) {
-        PyErr_Format(PyExc_ValueError, "%d is not in bitarray", vi);
-        return NULL;
-    }
+    if (i < 0)
+        return PyErr_Format(PyExc_ValueError, "%d is not in bitarray", vi);
+
     return PyLong_FromSsize_t(i);
 }
 
@@ -788,11 +799,11 @@ bitarray_extend(bitarrayobject *self, PyObject *obj)
 }
 
 PyDoc_STRVAR(extend_doc,
-"extend(iterable or string, /)\n\
+"extend(iterable, /)\n\
 \n\
 Extend bitarray by appending the truth value of each element given\n\
 by iterable.  If a string is provided, each `0` and `1` are appended\n\
-as bits.");
+as bits (whitespace is ignored).");
 
 
 static PyObject *
@@ -802,7 +813,7 @@ bitarray_search(bitarrayobject *self, PyObject *args)
     PyObject *item = NULL, *x;
     Py_ssize_t limit = -1;
     bitarrayobject *xa;
-    Py_ssize_t p;
+    Py_ssize_t p = 0;
 
     if (!PyArg_ParseTuple(args, "O|n:search", &x, &limit))
         return NULL;
@@ -822,13 +833,8 @@ bitarray_search(bitarrayobject *self, PyObject *args)
     if (xa->nbits > self->nbits || limit == 0)
         return list;
 
-    p = 0;
-    while (1) {
-        p = search(self, xa, p);
-        if (p < 0)
-            break;
-        item = PyLong_FromSsize_t(p);
-        p++;
+    while ((p = search(self, xa, p)) >= 0) {
+        item = PyLong_FromSsize_t(p++);
         if (item == NULL || PyList_Append(list, item) < 0) {
             Py_XDECREF(item);
             Py_XDECREF(list);
@@ -910,7 +916,7 @@ bitarray_all(bitarrayobject *self)
 PyDoc_STRVAR(all_doc,
 "all() -> bool\n\
 \n\
-Returns True when all bits in the array are True.");
+Return True when all bits in the array are True.");
 
 
 static PyObject *
@@ -922,7 +928,7 @@ bitarray_any(bitarrayobject *self)
 PyDoc_STRVAR(any_doc,
 "any() -> bool\n\
 \n\
-Returns True when any bit in the array is True.");
+Return True when any bit in the array is True.");
 
 
 static PyObject *
@@ -938,14 +944,14 @@ bitarray_reduce(bitarrayobject *self)
         dict = Py_None;
         Py_INCREF(dict);
     }
-    /* the first byte indicates the number of unused bits at the end, and
-       the rest of the bytes consist of the raw binary data */
     data = (char *) PyMem_Malloc(nbytes + 1);
     if (data == NULL) {
         PyErr_NoMemory();
         goto error;
     }
-    data[0] = (char) setunused(self);
+    /* first byte contains the number of unused bits */
+    *data = (char) setunused(self);
+    /* remaining bytes contain buffer */
     memcpy(data + 1, self->ob_item, (size_t) nbytes);
     repr = PyBytes_FromStringAndSize(data, nbytes + 1);
     if (repr == NULL)
@@ -953,7 +959,7 @@ bitarray_reduce(bitarrayobject *self)
     PyMem_Free((void *) data);
     result = Py_BuildValue("O(Os)O", Py_TYPE(self),
                            repr, ENDIAN_STR(self), dict);
-error:
+ error:
     Py_DECREF(dict);
     Py_XDECREF(repr);
     return result;
@@ -962,6 +968,36 @@ error:
 PyDoc_STRVAR(reduce_doc, "state information for pickling");
 
 
+/* The head byte % 8 specifies the number of unused bits (in last buffer
+   byte), the remaining bytes consist of the buffer itself */
+static PyObject *
+unpickle(PyTypeObject *type, PyObject *bytes, int endian)
+{
+    PyObject *res;
+    Py_ssize_t nbytes;
+    unsigned char head;
+    char *data;
+
+    assert(PyBytes_Check(bytes));
+    nbytes = PyBytes_GET_SIZE(bytes);
+    assert(nbytes > 0);
+    data = PyBytes_AS_STRING(bytes);
+    head = *data;
+
+    if (nbytes == 1 && head % 8)
+        return PyErr_Format(PyExc_ValueError,
+                            "invalid header byte 0x%02x", head);
+
+    res = newbitarrayobject(type,
+                            BITS(nbytes - 1) - ((Py_ssize_t) (head % 8)),
+                            endian);
+    if (res == NULL)
+        return NULL;
+    memcpy(((bitarrayobject *) res)->ob_item, data + 1, (size_t) nbytes - 1);
+    return res;
+}
+
+
 static PyObject *
 bitarray_reverse(bitarrayobject *self)
 {
@@ -1011,8 +1047,8 @@ bitarray_fill(bitarrayobject *self)
 PyDoc_STRVAR(fill_doc,
 "fill() -> int\n\
 \n\
-Adds zeros to the end of the bitarray, such that the length of the bitarray\n\
-will be a multiple of 8.  Returns the number of bits added (0..7).");
+Add zeros to the end of the bitarray, such that the length of the bitarray\n\
+will be a multiple of 8, and return the number of bits added (0..7).");
 
 
 static PyObject *
@@ -1040,7 +1076,7 @@ bitarray_invert(bitarrayobject *self, PyObject *args)
 }
 
 PyDoc_STRVAR(invert_doc,
-"invert(index=<all bits>)\n\
+"invert(index=<all bits>, /)\n\
 \n\
 Invert all bits in the array (in-place).\n\
 When the optional `index` is given, only invert the single bit at index.");
@@ -1049,28 +1085,7 @@ When the optional `index` is given, only invert the single bit at index.");
 static PyObject *
 bitarray_bytereverse(bitarrayobject *self)
 {
-    static char trans[256];
-    static int setup = 0;
-    Py_ssize_t i;
-
-    if (!setup) {
-        /* setup translation table, which maps each byte to it's reversed:
-           trans = {0, 128, 64, 192, 32, 160, ..., 255} */
-        int j, k;
-
-        for (k = 0; k < 256; k++) {
-            trans[k] = 0x00;
-            for (j = 0; j < 8; j++)
-                if (1 << (7 - j) & k)
-                    trans[k] |= 1 << j;
-        }
-        setup = 1;
-    }
-
-    setunused(self);
-    for (i = 0; i < Py_SIZE(self); i++)
-        self->ob_item[i] = trans[(unsigned char) self->ob_item[i]];
-
+    bytereverse(self, 0, Py_SIZE(self));
     Py_RETURN_NONE;
 }
 
@@ -1195,7 +1210,7 @@ bitarray_frombytes(bitarrayobject *self, PyObject *bytes)
         return NULL;
 
     memcpy(self->ob_item + (Py_SIZE(self) - nbytes),
-           PyBytes_AsString(bytes), (size_t) nbytes);
+           PyBytes_AS_STRING(bytes), (size_t) nbytes);
 
     if (delete_n(self, t, p) < 0)
         return NULL;
@@ -1272,7 +1287,7 @@ PyDoc_STRVAR(fromfile_doc,
 \n\
 Extend bitarray with up to n bytes read from the file object f.\n\
 When n is omitted or negative, reads all data until EOF.\n\
-When n is provided and positions but exceeds the data available,\n\
+When n is provided and positive but exceeds the data available,\n\
 EOFError is raised (but the available data is still read and appended.");
 
 
@@ -1355,7 +1370,7 @@ bitarray_pack(bitarrayobject *self, PyObject *bytes)
     if (resize(self, self->nbits + nbytes) < 0)
         return NULL;
 
-    data = PyBytes_AsString(bytes);
+    data = PyBytes_AS_STRING(bytes);
     for (i = 0; i < nbytes; i++)
         setbit(self, self->nbits - nbytes + i, data[i] ? 1 : 0);
 
@@ -1385,17 +1400,16 @@ bitarray_repr(bitarrayobject *self)
         return Py_BuildValue("s", "bitarray()");
 
     strsize = self->nbits + 12;  /* 12 is the length of "bitarray('')" */
-    if (strsize > PY_SSIZE_T_MAX / 8) {
+    if (strsize > PY_SSIZE_T_MAX) {
         PyErr_SetString(PyExc_OverflowError,
                         "bitarray too large to represent");
         return NULL;
     }
 
     str = (char *) PyMem_Malloc(strsize);
-    if (str == NULL) {
-        PyErr_NoMemory();
-        return NULL;
-    }
+    if (str == NULL)
+        return PyErr_NoMemory();
+
     /* add "bitarray('......')" to str */
     strcpy(str, "bitarray('"); /* has length 10 */
     /* don't use strcpy here, as this would add an extra null byte */
@@ -1480,10 +1494,9 @@ bitarray_remove(bitarrayobject *self, PyObject *v)
         return NULL;
 
     i = findfirst(self, vi, 0, self->nbits);
-    if (i < 0) {
-        PyErr_Format(PyExc_ValueError, "%d not in bitarray", vi);
-        return NULL;
-    }
+    if (i < 0)
+        return PyErr_Format(PyExc_ValueError, "%d not in bitarray", vi);
+
     if (delete_n(self, i, 1) < 0)
         return NULL;
     Py_RETURN_NONE;
@@ -1676,8 +1689,10 @@ bitarray_subscr(bitarrayobject *self, PyObject *item)
         }
         return res;
     }
-    PyErr_SetString(PyExc_TypeError, "index or slice expected");
-    return NULL;
+
+    return PyErr_Format(PyExc_TypeError,
+                        "bitarray indices must be integers or slices, not %s",
+                        Py_TYPE(item)->tp_name);
 }
 
 /* The following functions (setslice_bitarray, setslice_bool and delslice)
@@ -1828,11 +1843,16 @@ bitarray_ass_subscr(bitarrayobject *self, PyObject* item, PyObject* value)
         if (IS_INT_OR_BOOL(value))
             return setslice_bool(self, item, value);
 
-        PyErr_SetString(PyExc_IndexError,
-                        "bitarray or bool expected for slice assignment");
+        PyErr_Format(PyExc_TypeError,
+                     "bitarray or bool expected for slice assignment, not %s",
+                     Py_TYPE(value)->tp_name);
         return -1;
     }
-    return -1;                  /* cannot happen */
+
+    PyErr_Format(PyExc_TypeError,
+                 "bitarray indices must be integers or slices, not %s",
+                 Py_TYPE(item)->tp_name);
+    return -1;
 }
 
 static PyMappingMethods bitarray_as_mapping = {
@@ -2012,7 +2032,8 @@ static int
 check_codedict(PyObject *codedict)
 {
     if (!PyDict_Check(codedict)) {
-        PyErr_SetString(PyExc_TypeError, "dict expected");
+        PyErr_Format(PyExc_TypeError, "dict expected, got %s",
+                     Py_TYPE(codedict)->tp_name);
         return -1;
     }
     if (PyDict_Size(codedict) == 0) {
@@ -2049,12 +2070,12 @@ bitarray_encode(bitarrayobject *self, PyObject *args)
         return NULL;
 
     iter = PyObject_GetIter(iterable);
-    if (iter == NULL) {
-        PyErr_SetString(PyExc_TypeError, "iterable object expected");
-        return NULL;
-    }
+    if (iter == NULL)
+        return PyErr_Format(PyExc_TypeError, "'%s' object is not iterable",
+                            Py_TYPE(iterable)->tp_name);
+
     /* extend self with the bitarrays from codedict */
-    while ((symbol = PyIter_Next(iter)) != NULL) {
+    while ((symbol = PyIter_Next(iter))) {
         value = PyDict_GetItem(codedict, symbol);
         Py_DECREF(symbol);
         if (value == NULL) {
@@ -2070,7 +2091,8 @@ bitarray_encode(bitarrayobject *self, PyObject *args)
     if (PyErr_Occurred())
         return NULL;
     Py_RETURN_NONE;
-error:
+
+ error:
     Py_DECREF(iter);
     return NULL;
 }
@@ -2235,7 +2257,8 @@ binode_to_dict(binode *nd, PyObject *dict, bitarrayobject *prefix)
         t = (bitarrayobject *) bitarray_copy(prefix);
         if (t == NULL)
             return -1;
-        resize(t, t->nbits + 1);
+        if (resize(t, t->nbits + 1) < 0)
+            return -1;
         setbit(t, t->nbits - 1, k);
         ret = binode_to_dict(nd->child[k], dict, t);
         Py_DECREF((PyObject *) t);
@@ -2434,7 +2457,7 @@ bitarray_decode(bitarrayobject *self, PyObject *obj)
 
         tree = binode_make_tree(obj);
         if (tree == NULL)
-            goto error;
+            return NULL;
     }
 
     list = PyList_New(0);
@@ -2451,7 +2474,7 @@ bitarray_decode(bitarrayobject *self, PyObject *obj)
         binode_delete(tree);
     return list;
 
-error:
+ error:
     if (!DecodeTree_Check(obj))
         binode_delete(tree);
     Py_XDECREF(list);
@@ -2477,9 +2500,6 @@ typedef struct {
 
 static PyTypeObject DecodeIter_Type;
 
-#define DecodeIter_Check(op)  PyObject_TypeCheck(op, &DecodeIter_Type)
-
-
 /* create a new initialized bitarray decode iterator object */
 static PyObject *
 bitarray_iterdecode(bitarrayobject *self, PyObject *obj)
@@ -2500,8 +2520,11 @@ bitarray_iterdecode(bitarrayobject *self, PyObject *obj)
     }
 
     it = PyObject_GC_New(decodeiterobject, &DecodeIter_Type);
-    if (it == NULL)
+    if (it == NULL) {
+        if (!DecodeTree_Check(obj))
+            binode_delete(tree);
         return NULL;
+    }
 
     Py_INCREF(self);
     it->bao = self;
@@ -2525,7 +2548,6 @@ decodeiter_next(decodeiterobject *it)
 {
     PyObject *symbol;
 
-    assert(DecodeIter_Check(it));
     symbol = binode_traverse(it->tree, it->bao, &(it->index));
     if (symbol == NULL)  /* stop iteration OR error occured */
         return NULL;
@@ -2601,8 +2623,6 @@ typedef struct {
 
 static PyTypeObject SearchIter_Type;
 
-#define SearchIter_Check(op)  PyObject_TypeCheck(op, &SearchIter_Type)
-
 /* create a new initialized bitarray search iterator object */
 static PyObject *
 bitarray_itersearch(bitarrayobject *self, PyObject *x)
@@ -2644,7 +2664,6 @@ searchiter_next(searchiterobject *it)
 {
     Py_ssize_t p;
 
-    assert(SearchIter_Check(it));
     p = search(it->bao, it->xa, it->p);
     if (p < 0)  /* no more positions -- stop iteration */
         return NULL;
@@ -2825,6 +2844,23 @@ endian_from_string(const char* string)
     return -1;
 }
 
+static PyObject *
+bitarray_from_index(PyTypeObject *type, PyObject *index, int endian)
+{
+    Py_ssize_t nbits;
+
+    assert(PyIndex_Check(index));
+    nbits = PyNumber_AsSsize_t(index, PyExc_IndexError);
+    if (nbits == -1 && PyErr_Occurred())
+        return NULL;
+
+    if (nbits < 0) {
+        PyErr_SetString(PyExc_ValueError,
+                        "cannot create bitarray with negative length");
+        return NULL;
+    }
+    return newbitarrayobject(type, nbits, endian);
+}
 
 static PyObject *
 bitarray_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
@@ -2847,72 +2883,30 @@ bitarray_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
     if (initial == NULL || initial == Py_None)
         return newbitarrayobject(type, 0, endian);
 
-    /* boolean */
+    /* bool */
     if (PyBool_Check(initial)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "cannot create bitarray from boolean");
+        PyErr_SetString(PyExc_TypeError, "cannot create bitarray from bool");
         return NULL;
     }
 
     /* index (a number) */
-    if (PyIndex_Check(initial)) {
-        Py_ssize_t nbits;
-
-        nbits = PyNumber_AsSsize_t(initial, PyExc_IndexError);
-        if (nbits == -1 && PyErr_Occurred())
-            return NULL;
-
-        if (nbits < 0) {
-            PyErr_SetString(PyExc_ValueError,
-                            "cannot create bitarray with negative length");
-            return NULL;
-        }
-        return newbitarrayobject(type, nbits, endian);
-    }
-
-    /* from bitarray itself */
-    if (bitarray_Check(initial)) {
-#define np  ((bitarrayobject *) initial)
-        res = newbitarrayobject(type, np->nbits,
-                                endian_str == NULL ? np->endian : endian);
-        if (res == NULL)
-            return NULL;
-        memcpy(((bitarrayobject *) res)->ob_item, np->ob_item,
-               (size_t) Py_SIZE(np));
-#undef np
-        return res;
-    }
+    if (PyIndex_Check(initial))
+        return bitarray_from_index(type, initial, endian);
 
     /* bytes (for pickling) */
-    if (PyBytes_Check(initial)) {
-        Py_ssize_t nbytes;
-        char *data;
+    if (PyBytes_Check(initial) && PyBytes_GET_SIZE(initial) > 0) {
+        unsigned char head = *PyBytes_AS_STRING(initial);
 
-        nbytes = PyBytes_Size(initial);
-        if (nbytes == 0)        /* no bytes */
-            return newbitarrayobject(type, 0, endian);
-
-        data = PyBytes_AsString(initial);
-        if (0 <= data[0] && data[0] < 8) {
-            /* when the first character is smaller than 8, it indicates the
-               number of unused bits at the end, and rest of the bytes
-               consist of the raw binary data */
-            if (nbytes == 1 && data[0] > 0) {
-                PyErr_Format(PyExc_ValueError,
-                             "did not expect 0x0%d", (int) data[0]);
-                return NULL;
-            }
-            res = newbitarrayobject(type,
-                                    BITS(nbytes - 1) - ((Py_ssize_t) data[0]),
-                                    endian);
-            if (res == NULL)
-                return NULL;
-            memcpy(((bitarrayobject *) res)->ob_item, data + 1,
-                   (size_t) nbytes - 1);
-            return res;
+        if (head < 32 && head % 16 < 8) {
+            if (endian_str == NULL)  /* no endianness given in argument */
+                endian = head / 16 ? ENDIAN_BIG : ENDIAN_LITTLE;
+            return unpickle(type, initial, endian);
         }
     }
 
+    if (bitarray_Check(initial) && endian_str == NULL)
+        endian = ((bitarrayobject *) initial)->endian;
+
     /* leave remaining type dispatch to the extend method */
     res = newbitarrayobject(type, 0, endian);
     if (res == NULL)
@@ -2960,17 +2954,12 @@ richcompare(PyObject *v, PyObject *w, int op)
         vi = GETBIT(va, i);
         wi = GETBIT(wa, i);
         if (vi != wi) {
-            /* we have an item that differs -- first, shortcut for EQ/NE */
-            if (op == Py_EQ)
-                Py_RETURN_FALSE;
-            if (op == Py_NE)
-                Py_RETURN_TRUE;
-            /* compare the final item using the proper operator */
+            /* we have an item that differs */
             switch (op) {
             case Py_LT: cmp = vi <  wi; break;
             case Py_LE: cmp = vi <= wi; break;
-            case Py_EQ: cmp = vi == wi; break;
-            case Py_NE: cmp = vi != wi; break;
+            case Py_EQ: Py_RETURN_FALSE;
+            case Py_NE: Py_RETURN_TRUE;
             case Py_GT: cmp = vi >  wi; break;
             case Py_GE: cmp = vi >= wi; break;
             default: return NULL;  /* cannot happen */
@@ -3004,8 +2993,6 @@ typedef struct {
 
 static PyTypeObject BitarrayIter_Type;
 
-#define BitarrayIter_Check(op)  PyObject_TypeCheck(op, &BitarrayIter_Type)
-
 /* create a new initialized bitarray iterator object, this object is
    returned when calling iter(a) */
 static PyObject *
@@ -3013,7 +3000,6 @@ bitarray_iter(bitarrayobject *self)
 {
     bitarrayiterobject *it;
 
-    assert(bitarray_Check(self));
     it = PyObject_GC_New(bitarrayiterobject, &BitarrayIter_Type);
     if (it == NULL)
         return NULL;
@@ -3030,7 +3016,6 @@ bitarrayiter_next(bitarrayiterobject *it)
 {
     long vi;
 
-    assert(BitarrayIter_Check(it));
     if (it->index < it->bao->nbits) {
         vi = GETBIT(it->bao, it->index);
         it->index++;
@@ -3281,8 +3266,8 @@ set_default_endian(PyObject *module, PyObject *args)
     if (!PyArg_ParseTuple(args, "s:_set_default_endian", &endian_str))
         return NULL;
 
-    /* As endian_from_string might return -1, we have to store its value
-       in a temporary variable before setting default_endian. */
+    /* As endian_from_string() might return -1, we have to store its value
+       in a temporary variable BEFORE setting default_endian. */
     tmp = endian_from_string(endian_str);
     if (tmp < 0)
         return NULL;
@@ -3319,11 +3304,12 @@ tuple(sizeof(void *),\n\
 
 
 static PyMethodDef module_functions[] = {
-    {"get_default_endian", (PyCFunction) get_default_endian, METH_NOARGS,
-                                                   get_default_endian_doc},
-    {"_set_default_endian", (PyCFunction) set_default_endian, METH_VARARGS,
-                                                   set_default_endian_doc},
-    {"_sysinfo",   (PyCFunction) sysinfo,    METH_NOARGS,  sysinfo_doc   },
+    {"get_default_endian",  (PyCFunction) get_default_endian,
+                                      METH_NOARGS,  get_default_endian_doc},
+    {"_set_default_endian", (PyCFunction) set_default_endian,
+                                      METH_VARARGS, set_default_endian_doc},
+    {"_sysinfo",            (PyCFunction) sysinfo,
+                                      METH_NOARGS,  sysinfo_doc           },
     {"eval_all_terms", (PyCFunction) eval_all_terms, METH_VARARGS | METH_KEYWORDS, eval_all_terms_doc },
     {NULL,         NULL}  /* sentinel */
 };
