@@ -1,10 +1,10 @@
 /*
-   Copyright (c) 2008 - 2021, Ilan Schnell
+   Copyright (c) 2008 - 2021, Ilan Schnell; All Rights Reserved
    bitarray is published under the PSF license.
 
    Author: Ilan Schnell
 */
-#define BITARRAY_VERSION  "1.6.4"
+#define BITARRAY_VERSION  "1.9.0"
 
 #if PY_MAJOR_VERSION >= 3
 #define IS_PY3K
@@ -35,7 +35,7 @@ typedef struct {
 #define ENDIAN_LITTLE  0
 #define ENDIAN_BIG     1
 
-/* returns the endianness string from bitarrayobject */
+/* the endianness string from bitarrayobject */
 #define ENDIAN_STR(o)  ((o)->endian == ENDIAN_LITTLE ? "little" : "big")
 
 /* number of bits that can be stored in given bytes */
@@ -50,7 +50,8 @@ typedef struct {
 /* ------------ low level access to bits in bitarrayobject ------------- */
 
 #ifndef NDEBUG
-static inline int GETBIT(bitarrayobject *self, Py_ssize_t i) {
+static inline int GETBIT(bitarrayobject *self, Py_ssize_t i)
+{
     assert(0 <= i && i < self->nbits);
     return ((self)->ob_item[(i) / 8] & BITMASK((self)->endian, i) ? 1 : 0);
 }
@@ -97,7 +98,7 @@ setunused(bitarrayobject *self)
     return res;
 }
 
-static unsigned char bitcount_lookup[256] = {
+static const unsigned char bitcount_lookup[256] = {
     0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
